import morepath

from onegov.chat import MessageCollection
from onegov.core.utils import normalize_for_url
from onegov.core.custom import json
from onegov.core.orm import as_selectable
from onegov.core.security import Public, Private
from onegov.org import _, OrgApp
from onegov.core.elements import Link, Intercooler, Confirm
from onegov.org.forms import TicketNoteForm
from onegov.org.forms import TicketChatMessageForm
from onegov.org.forms import InternalTicketChatMessageForm
from onegov.org.layout import DefaultLayout
from onegov.org.layout import TicketLayout
from onegov.org.layout import TicketNoteLayout
from onegov.org.layout import TicketsLayout
from onegov.org.layout import TicketChatMessageLayout
from onegov.org.mail import send_ticket_mail
from onegov.org.models import TicketChatMessage, TicketMessage, TicketNote
from onegov.org.views.message import view_messages_feed
from onegov.ticket import handlers as ticket_handlers
from onegov.ticket import Ticket, TicketCollection
from onegov.ticket.errors import InvalidStateChange
from onegov.user import User, UserCollection
from sqlalchemy import select


@OrgApp.html(model=Ticket, template='ticket.pt', permission=Private)
def view_ticket(self, request):

    handler = self.handler

    if handler.deleted:
        summary = self.snapshot.get('summary')
    else:
        # XXX this is very to do here, much harder when the ticket is updated
        # because there's no good link to the ticket at that point - so when
        # viewing the ticket we commit the sin of possibly changing data in a
        # GET request.
        handler.refresh()
        summary = handler.get_summary(request)

    if handler.payment:
        handler.payment.sync()

    messages = MessageCollection(
        request.session,
        channel_id=self.number
    )

    stmt = as_selectable("""
        SELECT
            channel_id,    -- Text
            SUM(
                CASE WHEN type = 'ticket_note' THEN
                    1 ELSE 0 END
            ) AS notes,    -- Integer

            SUM(CASE WHEN type = 'ticket_chat' THEN
                    CASE WHEN meta->>'origin' = 'internal' THEN 1
                    ELSE 0
                END ELSE 0 END
            ) AS internal, -- Integer

            SUM(CASE WHEN type = 'ticket_chat' THEN
                    CASE WHEN meta->>'origin' = 'external' THEN 1
                    ELSE 0
                END ELSE 0 END
            ) AS external  -- Integer

        FROM messages
        WHERE type IN ('ticket_note', 'ticket_chat')
        GROUP BY channel_id
    """)

    counts = request.session.execute(
        select(stmt.c).where(stmt.c.channel_id == self.number)).first()

    # if we have a payment, show the payment button
    layout = TicketLayout(self, request)
    payment_button = None

    if self.state == 'pending':
        payment = handler.payment

        if payment and payment.source == 'manual':
            payment_button = manual_payment_button(payment, layout)

        if payment and payment.source == 'stripe_connect':
            payment_button = stripe_payment_button(payment, layout)

    return {
        'title': self.number,
        'layout': layout,
        'ticket': self,
        'summary': summary,
        'deleted': handler.deleted,
        'handler': handler,
        'payment_button': payment_button,
        'counts': counts,
        'feed_data': json.dumps(
            view_messages_feed(messages, request)
        ),
    }


def manual_payment_button(payment, layout):
    if payment.state == 'open':
        return Link(
            text=_("Mark as paid"),
            url=layout.csrf_protected_url(
                layout.request.link(payment, 'mark-as-paid'),
            ),
            attrs={'class': 'mark-as-paid button small secondary'},
            traits=(
                Intercooler(
                    request_method='POST',
                    redirect_after=layout.request.url,
                ),
            )
        )

    return Link(
        text=_("Mark as unpaid"),
        url=layout.csrf_protected_url(
            layout.request.link(payment, 'mark-as-unpaid'),
        ),
        attrs={'class': 'mark-as-unpaid button small secondary'},
        traits=(
            Intercooler(
                request_method='POST',
                redirect_after=layout.request.url,
            ),
        )
    )


def stripe_payment_button(payment, layout):
    if payment.state == 'open':
        return Link(
            text=_("Capture Payment"),
            url=layout.csrf_protected_url(
                layout.request.link(payment, 'capture')
            ),
            attrs={'class': 'payment-capture button small secondary'},
            traits=(
                Confirm(
                    _("Do you really want capture the payment?"),
                    _(
                        "This usually happens automatically, so there is "
                        "no reason not do capture the payment."
                    ),
                    _("Capture payment"),
                    _("Cancel")
                ),
                Intercooler(
                    request_method='POST',
                    redirect_after=layout.request.url
                ),
            )
        )

    if payment.state == 'paid':
        amount = '{:02f} {}'.format(payment.amount, payment.currency)

        return Link(
            text=_("Refund Payment"),
            url=layout.csrf_protected_url(
                layout.request.link(payment, 'refund')
            ),
            attrs={'class': 'payment-refund button small secondary'},
            traits=(
                Confirm(
                    _("Do you really want to refund ${amount}?", mapping={
                        'amount': amount
                    }),
                    _("This cannot be undone."),
                    _("Refund ${amount}", mapping={
                        'amount': amount
                    }),
                    _("Cancel")
                ),
                Intercooler(
                    request_method='POST',
                    redirect_after=layout.request.url
                )
            )
        )


def send_email_if_enabled(ticket, request, template, subject):
    email = ticket.snapshot.get('email') or ticket.handler.email

    send_ticket_mail(
        request=request,
        template=template,
        subject=subject,
        receivers=(email, ),
        ticket=ticket
    )


def last_internal_message(session, ticket_number):
    messages = MessageCollection(
        session,
        type='ticket_chat',
        channel_id=ticket_number,
        load='newer-first'
    )

    return messages.query()\
        .filter(TicketChatMessage.meta['origin'].astext == 'internal')\
        .first()


def send_chat_message_email_if_enabled(ticket, request, message, origin):
    assert origin in ('internal', 'external')

    messages = MessageCollection(
        request.session,
        channel_id=ticket.number,
        type='ticket_chat')

    if origin == 'internal':

        # if the messages is sent to the outside, we always send an e-mail
        receivers = (ticket.handler.email, )
        reply_to = request.current_username

    else:
        # if the message is sent to the inside, we check the setting on the
        # last message sent to the outside in this ticket - if none exists,
        # we do not notify
        last_internal = last_internal_message(request.session, ticket.number)

        if not last_internal or not last_internal.meta.get('notify'):
            return

        receivers = (last_internal.owner, )
        reply_to = None  # default reply-to given by the application

    # we show the previous messages by going back until we find a message
    # that is not from the same author as the new message (this should usually
    # be the next message, but might include multiple, if someone sent a bunch
    # of messages in succession without getting a reply)
    #
    # note that the resulting thread has to be reversed for the mail template
    def thread():
        messages.older_than = message.id
        messages.load = 'newer-first'

        for m in messages.query():
            yield m

            if m.owner != message.owner:
                break

    send_ticket_mail(
        request=request,
        template='mail_ticket_chat_message.pt',
        subject=_("Your ticket has a new message"),
        content={
            'model': ticket,
            'message': message,
            'thread': tuple(reversed(list(thread()))),
        },
        ticket=ticket,
        receivers=receivers,
        reply_to=reply_to,
        force=True
    )


@OrgApp.form(
    model=Ticket, name='note', permission=Private,
    template='ticket_note_form.pt', form=TicketNoteForm
)
def handle_new_note(self, request, form):

    if form.submitted(request):
        TicketNote.create(self, request, form.text.data, form.file.create())
        request.success(_("Your note was added"))
        return request.redirect(request.link(self))

    return {
        'title': _("New Note"),
        'layout': TicketNoteLayout(self, request, _("New Note")),
        'form': form,
        'hint': 'default'
    }


@OrgApp.view(model=TicketNote, permission=Private)
def view_ticket_note(self, request):
    return request.redirect(request.link(self.ticket))


@OrgApp.view(model=TicketNote, permission=Private, request_method='DELETE')
def delete_ticket_note(self, request):
    request.assert_valid_csrf_token()

    # force a change of the ticket to make sure that it gets reindexed
    self.ticket.force_update()

    request.session.delete(self)
    request.success(_("The note was deleted"))


@OrgApp.form(
    model=TicketNote, name='edit', permission=Private,
    template='ticket_note_form.pt', form=TicketNoteForm
)
def handle_edit_note(self, request, form):
    if form.submitted(request):
        form.populate_obj(self)
        self.owner = request.current_username

        # force a change of the ticket to make sure that it gets reindexed
        self.ticket.force_update()

        request.success(_("Your changes were saved"))
        return request.redirect(request.link(self.ticket))

    elif not request.POST:
        form.process(obj=self)

    return {
        'title': _("Edit Note"),
        'layout': TicketNoteLayout(self.ticket, request, _("New Note")),
        'form': form,
        'hint': self.owner != request.current_username and 'owner'
    }


@OrgApp.view(model=Ticket, name='accept', permission=Private)
def accept_ticket(self, request):
    user = UserCollection(request.session).by_username(
        request.identity.userid)

    was_pending = self.state == 'open'

    try:
        self.accept_ticket(user)
    except InvalidStateChange:
        request.alert(_("The ticket cannot be accepted because it's not open"))
    else:
        if was_pending:
            TicketMessage.create(self, request, 'accepted')
            request.success(_("You have accepted ticket ${number}", mapping={
                'number': self.number
            }))

    return morepath.redirect(request.link(self))


@OrgApp.view(model=Ticket, name='close', permission=Private)
def close_ticket(self, request):

    was_pending = self.state == 'pending'

    try:
        self.close_ticket()
    except InvalidStateChange:
        request.alert(
            _("The ticket cannot be closed because it's not pending")
        )
    else:
        if was_pending:
            TicketMessage.create(self, request, 'closed')
            request.success(_("You have closed ticket ${number}", mapping={
                'number': self.number
            }))

            send_email_if_enabled(
                ticket=self,
                request=request,
                template='mail_ticket_closed.pt',
                subject=_("Your ticket has been closed")
            )

    return morepath.redirect(
        request.link(TicketCollection(request.session)))


@OrgApp.view(model=Ticket, name='reopen', permission=Private)
def reopen_ticket(self, request):
    user = UserCollection(request.session).by_username(
        request.identity.userid)

    was_closed = self.state == 'closed'

    try:
        self.reopen_ticket(user)
    except InvalidStateChange:
        request.alert(
            _("The ticket cannot be re-opened because it's not closed")
        )
    else:
        if was_closed:
            TicketMessage.create(self, request, 'reopened')
            request.success(_("You have reopened ticket ${number}", mapping={
                'number': self.number
            }))

            send_email_if_enabled(
                ticket=self,
                request=request,
                template='mail_ticket_reopened.pt',
                subject=_("Your ticket has been reopened")
            )

    return morepath.redirect(request.link(self))


@OrgApp.view(model=Ticket, name='mute', permission=Private)
def mute_ticket(self, request):
    self.muted = True

    TicketMessage.create(self, request, 'muted')
    request.success(
        _("You have disabled e-mails for ticket ${number}", mapping={
            'number': self.number
        }))

    return morepath.redirect(request.link(self))


@OrgApp.view(model=Ticket, name='unmute', permission=Private)
def unmute_ticket(self, request):
    self.muted = False

    TicketMessage.create(self, request, 'unmuted')
    request.success(
        _("You have enabled e-mails for ticket ${number}", mapping={
            'number': self.number
        }))

    return morepath.redirect(request.link(self))


@OrgApp.form(model=Ticket, name='message-to-submitter', permission=Private,
             form=InternalTicketChatMessageForm, template='form.pt')
def message_to_submitter(self, request, form):
    recipient = self.handler.email

    if form.submitted(request):
        if self.state == 'closed':
            request.alert(_("The ticket has already been closed"))
        else:
            message = TicketChatMessage.create(
                self, request,
                text=form.text.data,
                owner=request.current_username,
                recipient=recipient,
                notify=form.notify.data,
                origin='internal')

            send_chat_message_email_if_enabled(
                self, request, message, origin='internal')

            request.success(_("Your message has been sent"))
            return morepath.redirect(request.link(self))
    elif not request.POST:
        # show the same notification setting as was selected with the
        # last internal message - otherwise default to False
        last_internal = last_internal_message(request.session, self.number)

        if last_internal:
            form.notify.data = last_internal.meta.get('notify', False)
        else:
            form.notify.data = False

    return {
        'title': _("New Message"),
        'layout': TicketChatMessageLayout(self, request),
        'form': form,
        'helptext': _(
            "The following message will be sent to ${address} and it will be "
            "recorded for future reference.", mapping={
                'address': recipient
            }
        )
    }


@OrgApp.form(model=Ticket, name='status', template='ticket_status.pt',
             permission=Public, form=TicketChatMessageForm)
def view_ticket_status(self, request, form):

    if self.state == 'open':
        title = _("Your request has been submitted")
    elif self.state == 'pending':
        title = _("Your request is currently pending")
    elif self.state == 'closed':
        title = _("Your request has been processed")
    else:
        raise NotImplementedError

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Ticket Status"), '#')
    ]

    if form.submitted(request):

        if self.state == 'closed':
            request.alert(_("The ticket has already been closed"))
        else:
            message = TicketChatMessage.create(
                self, request,
                text=form.text.data,
                owner=self.handler.email,
                origin='external')

            send_chat_message_email_if_enabled(
                self, request, message, origin='external')

            request.success(_("Your message has been received"))
            return morepath.redirect(request.link(self, 'status'))

    messages = MessageCollection(
        request.session,
        channel_id=self.number,
        type=request.app.settings.org.public_ticket_messages
    )

    return {
        'title': title,
        'layout': layout,
        'ticket': self,
        'feed_data': messages and json.dumps(
            view_messages_feed(messages, request)
        ) or None,
        'form': form
    }


@OrgApp.html(model=TicketCollection, template='tickets.pt',
             permission=Private)
def view_tickets(self, request):
    query = as_selectable("""
        SELECT
            handler_code,                         -- Text
            ARRAY_AGG(DISTINCT "group") AS groups -- ARRAY(Text)
        FROM tickets GROUP BY handler_code
    """)

    groups = {
        r.handler_code: r.groups
        for r in request.session.execute(select(query.c))
    }

    for handler in groups:
        groups[handler].sort(key=lambda g: normalize_for_url(g))

    def get_filters():
        states = (
            ('open', _("Open")),
            ('pending', _("Pending")),
            ('closed', _("Closed")),
            ('all', _("All"))
        )

        for id, text in states:
            yield Link(
                text=text,
                url=request.link(self.for_state(id)),
                active=self.state == id,
                attrs={'class': 'ticket-filter-' + id}
            )

    def get_handlers():
        nonlocal groups

        handlers = []

        for key, handler in ticket_handlers.registry.items():
            if key in groups:
                handlers.append(
                    (key, request.translate(handler.handler_title)))

        handlers.sort(key=lambda item: item[1])
        handlers.insert(0, ('ALL', _("All Tickets")))

        for id, text in handlers:
            grouplinks = id != 'ALL' and tuple(get_groups(id))
            parent = grouplinks and len(grouplinks) > 1
            classes = parent and (id + '-link', 'is-parent') or (id + '-link',)

            yield Link(
                text=text,
                url=request.link(self.for_handler(id).for_group(None)),
                active=self.handler == id and self.group is None,
                attrs={'class': ' '.join(classes)}
            )

            if parent:
                yield from grouplinks

    def get_owners():

        users = UserCollection(request.session)
        users = users.by_roles(*request.app.settings.org.ticket_manager_roles)
        users = users.order_by(User.title)

        yield Link(
            text=_("All Users"),
            url=request.link(self.for_owner('*')),
            active=self.owner == '*'
        )

        for user in users:
            yield Link(
                text=user.title,
                url=request.link(self.for_owner(user.id)),
                active=self.owner == user.id.hex,
                model=user
            )

    def get_groups(handler):
        base = self.for_handler(handler)

        for group in groups[handler]:
            yield Link(
                text=group,
                url=request.link(base.for_group(group)),
                active=self.handler == handler and self.group == group,
                attrs={'class': ' '.join(
                    (handler + '-sub-link', 'ticket-group-filter')
                )}
            )

    if self.state == 'open':
        tickets_title = _("Open Tickets")
    elif self.state == 'pending':
        tickets_title = _("Pending Tickets")
    elif self.state == 'closed':
        tickets_title = _("Closed Tickets")
    elif self.state == 'all':
        tickets_title = _("All Tickets")
    else:
        raise NotImplementedError

    handlers = tuple(get_handlers())
    owners = tuple(get_owners())
    filters = tuple(get_filters())

    handler = next((h for h in handlers if h.active), None)
    owner = next((o for o in owners if o.active), None)

    return {
        'title': _("Tickets"),
        'layout': TicketsLayout(self, request),
        'tickets': self.batch,
        'filters': filters,
        'handlers': handlers,
        'owners': owners,
        'tickets_title': tickets_title,
        'tickets_state': self.state,
        'has_handler_filter': self.handler != 'ALL',
        'has_owner_filter': self.owner != '*',
        'handler': handler,
        'owner': owner
    }



""" The settings view, defining things like the logo or color of the org. """

from copy import copy
from dectate import Query
from onegov.core.security import Secret
from onegov.core.templates import render_macro
from onegov.org import _
from onegov.org.app import OrgApp
from onegov.org.forms import AnalyticsSettingsForm
from onegov.org.forms import FooterSettingsForm
from onegov.org.forms import GeneralSettingsForm
from onegov.org.forms import HolidaySettingsForm
from onegov.org.forms import HomepageSettingsForm
from onegov.org.forms import MapSettingsForm
from onegov.org.forms import ModuleSettingsForm
from onegov.org.layout import DefaultLayout
from onegov.org.layout import SettingsLayout
from onegov.org.models import Organisation
from onegov.org.models import SwissHolidays


@OrgApp.html(
    model=Organisation, name='settings', template='settings.pt',
    permission=Secret)
def view_settings(self, request):
    layout = SettingsLayout(self, request)

    def query_settings():
        q = Query('view').filter(model=Organisation)

        for action, fn in q(request.app):
            if 'setting' in action.predicates:
                setting = copy(action.predicates)
                setting['title'] = setting['setting']
                setting['link'] = request.link(self, name=setting['name'])

                yield setting

    settings = list(query_settings())
    settings.sort(key=lambda s: s.get('order', 0))

    return {
        'layout': layout,
        'title': _("Settings"),
        'settings': settings
    }


def handle_generic_settings(self, request, form, title):
    layout = SettingsLayout(self, request, title)

    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return request.redirect(request.link(self, name='settings'))
    elif request.method == 'GET':
        form.process(obj=self)

    return {
        'layout': layout,
        'title': title,
        'form': form
    }


@OrgApp.form(
    model=Organisation, name='general-settings', template='form.pt',
    permission=Secret, form=GeneralSettingsForm, setting=_("General"),
    icon='fa-sliders', order=-1000)
def handle_general_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("General"))


@OrgApp.form(
    model=Organisation, name='homepage-settings', template='form.pt',
    permission=Secret, form=HomepageSettingsForm, setting=_("Homepage"),
    icon='fa-home', order=-900)
def handle_homepage_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("Homepage"))


@OrgApp.form(
    model=Organisation, name='footer-settings', template='form.pt',
    permission=Secret, form=FooterSettingsForm, setting=_("Footer"),
    icon='fa-window-minimize', order=-800)
def handle_footer_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("Footer"))


@OrgApp.form(
    model=Organisation, name='module-settings', template='form.pt',
    permission=Secret, form=ModuleSettingsForm, setting=_("Modules"),
    icon='fa-sitemap', order=-700)
def handle_module_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("Modules"))


@OrgApp.form(
    model=Organisation, name='map-settings', template='form.pt',
    permission=Secret, form=MapSettingsForm, setting=_("Map"),
    icon='fa-map-marker', order=-700)
def handle_map_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("Map"))


@OrgApp.form(
    model=Organisation, name='analytics-settings', template='form.pt',
    permission=Secret, form=AnalyticsSettingsForm, setting=_("Analytics"),
    icon='fa-line-chart ', order=-600)
def handle_analytics_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("Analytics"))


@OrgApp.form(
    model=Organisation, name='holiday-settings', template='form.pt',
    permission=Secret, form=HolidaySettingsForm, setting=_("Holidays"),
    icon='fa-calendar-o', order=-500)
def handle_holiday_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("Holidays"))


@OrgApp.form(model=Organisation, name='holiday-settings-preview',
             permission=Secret, form=HolidaySettingsForm)
def preview_holiday_settings(self, request, form):
    layout = DefaultLayout(self, request)

    if form.submitted(request):
        holidays = SwissHolidays(
            cantons=form.holiday_settings.get('cantons', ()),
            other=form.holiday_settings.get('other', ())
        )
    else:
        holidays = SwissHolidays(
            cantons=form.holiday_settings.get('cantons', ())
        )

    if not holidays.all(layout.today().year):
        msg = request.translate(_("No holidays defined"))
        return f'<i class="holidays">{msg}</i>'

    return render_macro(
        layout.macros['holidays'],
        request,
        {
            'holidays': holidays,
            'layout': layout,
            'year': layout.today().year,
        }
    )



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.upgrade import upgrade_task
from onegov.core.utils import normalize_for_url
from onegov.form import FormDefinition
from onegov.org.models import Organisation, Topic, News, ExtendedDirectory
from onegov.org.utils import annotate_html
from onegov.reservation import Resource


@upgrade_task('Move from town to organisation', always_run=True)
def move_town_to_organisation(context):

    # if the organisations table does not exist (other modules), skip
    has_organisations = context.app.session_manager.session().execute("""
        select exists (select 1 from information_schema.tables
        where table_schema='{}' and table_name='organisations')
    """.format(context.app.schema)).scalar()

    has_towns = context.app.session_manager.session().execute("""
        select exists (select 1 from information_schema.tables
        where table_schema='{}' and table_name='towns')
    """.format(context.app.schema)).scalar()

    if not has_organisations or not has_towns:
        return False

    # this needs to be an always-run task for now, since on first install the
    # module is deteced as new and it is assumed that no tasks need to be run
    # as a result
    if context.session.query(Organisation).first():
        return False

    context.operations.drop_table('organisations')
    context.operations.rename_table('towns', 'organisations')


@upgrade_task('Add image-dimensions to html')
def add_image_dimensions_to_html(context):
    for topic in context.session.query(Topic):
        topic.text = annotate_html(topic.text, context.request)

    for news in context.session.query(News):
        news.text = annotate_html(news.text, context.request)

    for resource in context.session.query(Resource):
        resource.text = annotate_html(resource.text, context.request)

    for form in context.session.query(FormDefinition):
        form.text = annotate_html(form.text, context.request)


@upgrade_task('Remove official notices table')
def remove_official_notices_table(context):
    # an incompatible release was accidentally left in, so we remove the
    # table in such instances, triggering a recreation on the next start

    if not context.has_table("official_notices"):
        return False

    session = context.app.session_manager.session()

    organisations_count = session.execute("select count(*) from organisations")
    if organisations_count.scalar() != 1:
        return False

    notices_count = session.execute("select count(*) from official_notices")
    if notices_count.scalar() != 0:
        return False

    context.operations.drop_table("official_notices")


@upgrade_task('Add new defaults to existing directories')
def add_new_defaults_to_existing_directories(context):
    for directory in context.session.query(ExtendedDirectory):
        directory.enable_submissions = False
        directory.price = 'free'
        directory.currency = 'CHF'


@upgrade_task('Migrate enable_map option')
def migrate_enable_map_option(context):
    for directory in context.session.query(ExtendedDirectory):
        directory.enable_map = directory.enable_map and 'everywhere' or 'no'


@upgrade_task('Fix directory order')
def fix_directory_order(context):
    for directory in context.session.query(ExtendedDirectory):
        directory.order = normalize_for_url(directory.title)


@upgrade_task('Add default redirect setting')
def add_default_recirect_setting(context):
    org = context.session.query(Organisation).first()

    if org:
        org.redirect_homepage_to = 'no'



from collections import defaultdict
from copy import copy
from onegov.core.crypto import random_password
from onegov.core.directives import query_form_class
from onegov.core.security import Secret
from onegov.core.templates import render_template
from onegov.form import merge_forms
from onegov.org import _, OrgApp
from onegov.org.forms import ManageUserForm, NewUserForm
from onegov.org.layout import DefaultMailLayout
from onegov.org.layout import UserLayout
from onegov.org.layout import UserManagementLayout
from onegov.core.elements import Link, LinkGroup
from onegov.ticket import TicketCollection, Ticket
from onegov.user import Auth, User, UserCollection
from onegov.user.errors import ExistingUserError
from onegov.user.forms import SignupLinkForm
from wtforms.validators import Optional


@OrgApp.html(model=UserCollection, template='usermanagement.pt',
             permission=Secret)
def view_usermanagement(self, request):
    """ Allows the management of organisation users. """

    layout = UserManagementLayout(self, request)

    users = defaultdict(list)

    for user in self.query().order_by(User.username).all():
        users[user.role].append(user)

    filters = {}

    filters['role'] = [
        Link(
            text=request.translate(title),
            active=value in self.filters.get('role', tuple()),
            url=request.link(self.for_filter(role=value))
        ) for title, value in (
            (_("Administrator"), 'admin'),
            (_("Editor"), 'editor'),
            (_("Member"), 'member'),
        )
    ]

    filters['active'] = [
        Link(
            text=request.translate(title),
            active=value in self.filters.get('active', tuple()),
            url=request.link(self.for_filter(active=value))
        ) for title, value in (
            (_("Active"), True),
            (_("Inactive"), False)
        )
    ]

    filters['tag'] = [
        Link(
            text=tag,
            active=tag in self.filters.get('tag', tuple()),
            url=request.link(self.for_filter(tag=tag))
        ) for tag in self.tags
    ]

    return {
        'layout': layout,
        'title': _("User Management"),
        'users': users,
        'filters': filters
    }


@OrgApp.form(
    model=UserCollection,
    template='signup_link.pt',
    permission=Secret,
    form=SignupLinkForm,
    name='signup-link')
def handle_create_signup_link(self, request, form):
    link = None

    if form.submitted(request):
        auth = Auth.from_app(request.app)
        auth.signup_token = form.signup_token(auth)

        link = request.link(auth, 'register')

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(_("New Signup Link"), '#'))
    layout.editbar_links = None

    return {
        'layout': layout,
        'title': _("New Signup Link"),
        'link': link,
        'form': form
    }


@OrgApp.html(model=User, template='user.pt', permission=Secret)
def view_user(self, request):
    """ Shows all objects owned by the given user. """

    layout = UserLayout(self, request)

    linkgroups = [
        fn(request, self) for fn in request.app.config.linkgroup_registry
    ]
    linkgroups.sort(key=lambda group: request.translate(group.title))

    return {
        'layout': layout,
        'title': self.title,
        'linkgroups': linkgroups
    }


@OrgApp.userlinks()
def ticket_links(request, user):
    tickets = TicketCollection(request.session).query()
    tickets = tickets.filter_by(user_id=user.id)
    tickets = tickets.order_by(Ticket.number)
    tickets = tickets.with_entities(
        Ticket.id, Ticket.number, Ticket.handler_code)

    return LinkGroup(
        title=_("Tickets"),
        links=[
            Link(
                ticket.number,
                request.class_link(Ticket, {
                    'handler_code': ticket.handler_code,
                    'id': ticket.id
                }),
            )
            for ticket in tickets
        ]
    )


def get_manage_user_form(self, request):
    userprofile_form = query_form_class(request, self, name='userprofile')
    assert userprofile_form

    class OptionalUserprofile(userprofile_form):

        hooked = False

        def submitted(self, request):
            # fields only present on the userprofile_form are made optional
            # to make sure that we can always change the active/inactive state
            # of the user and the role the user has
            if not self.hooked:
                for name, field in self._fields.items():
                    if not hasattr(userprofile_form, name):
                        continue

                    if not field.validators:
                        continue

                    # be careful not to change the class itself
                    field.validators = copy(field.validators)
                    field.validators.insert(0, Optional())

                self.hooked = True

            return super().submitted(request)

    return merge_forms(ManageUserForm, OptionalUserprofile)


@OrgApp.form(model=User, template='form.pt', form=get_manage_user_form,
             permission=Secret, name='edit')
def handle_manage_user(self, request, form):

    # XXX the manage user form doesn't have access to the username
    # because it can't be edited, so we need to inject it here
    # for validation purposes (check for a unique yubikey)
    form.current_username = self.username

    if not request.app.enable_yubikey:
        form.delete_field('yubikey')

    if form.submitted(request):
        form.populate_obj(self)
        request.success(_("Your changes were saved"))

        return request.redirect(request.class_link(UserCollection))

    elif not request.POST:
        form.process(obj=self)

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(self.username, '#'))

    return {
        'layout': layout,
        'title': self.username,
        'form': form
    }


@OrgApp.form(model=UserCollection, template='newuser.pt', form=NewUserForm,
             name='new', permission=Secret)
def handle_new_user(self, request, form):

    if not request.app.enable_yubikey:
        form.delete_field('yubikey')

    layout = UserManagementLayout(self, request)
    layout.breadcrumbs.append(Link(_("New User"), '#'))
    layout.editbar_links = None

    if form.submitted(request):
        password = random_password()

        if form.data.get('yubikey'):
            second_factor = {
                'type': 'yubikey',
                'data': form.data['yubikey'][:12]
            }
        else:
            second_factor = None

        try:
            user = self.add(
                username=form.username.data,
                password=password,
                role=form.role.data,
                active=form.active,
                second_factor=second_factor
            )
        except ExistingUserError:
            form.username.errors.append(
                _("A user with this e-mail address already exists"))
        else:
            if form.send_activation_email.data:
                subject = request.translate(
                    _("An account was created for you")
                )

                content = render_template('mail_new_user.pt', request, {
                    'user': user,
                    'org': request.app.org,
                    'layout': DefaultMailLayout(user, request),
                    'title': subject
                })

                request.app.send_transactional_email(
                    subject=subject,
                    receivers=(user.username, ),
                    content=content,
                )

            request.info(_("The user was created successfully"))

            return {
                'layout': layout,
                'title': _("New User"),
                'username': form.username.data,
                'password': password,
                'sent_email': form.send_activation_email.data
            }

    return {
        'layout': layout,
        'title': _("New User"),
        'form': form,
        'password': None,
        'sent_email': False
    }



""" Renders and handles defined forms, turning them into submissions. """

import morepath

from onegov.core.security import Public, Private
from onegov.ticket import TicketCollection
from onegov.form import (
    FormCollection,
    FormDefinition,
    PendingFormSubmission,
    CompleteFormSubmission
)
from onegov.org import _, OrgApp
from onegov.org.layout import FormSubmissionLayout
from onegov.org.mail import send_ticket_mail
from onegov.org.models import TicketMessage, SubmissionMessage
from onegov.pay import Price
from purl import URL


def copy_query(request, url, fields):
    url = URL(url)

    for field in fields:
        if field in request.GET:
            url = url.query_param(field, request.GET[field])

    return url.as_string()


def get_price(request, form, submission):
    total = form.total()

    if 'price' in submission.meta:
        if total is not None:
            total += Price(**submission.meta['price'])
        else:
            total = Price(**submission.meta['price'])

    return request.app.adjust_price(total)


def get_hints(layout, window):
    if not window:
        return

    if window.in_the_past:
        yield 'stop', _("The registration has ended")
    elif not window.enabled:
        yield 'stop', _("The registration is closed")

    if window.enabled and window.in_the_future:
        yield 'date', _("The registration opens on ${day}, ${date}", mapping={
            'day': layout.format_date(window.start, 'weekday_long'),
            'date': layout.format_date(window.start, 'date_long')
        })

    if window.enabled and window.in_the_present:
        yield 'date', _("The registration closes on ${day}, ${date}", mapping={
            'day': layout.format_date(window.end, 'weekday_long'),
            'date': layout.format_date(window.end, 'date_long')
        })

        if window.limit and window.overflow:
            yield 'count', _("There's a limit of ${count} attendees", mapping={
                'count': window.limit
            })

        if window.limit and not window.overflow:
            spots = window.available_spots

            if spots == 0:
                yield 'stop', _("There are no spots left")
            elif spots == 1:
                yield 'count', _("There is one spot left")
            else:
                yield 'count', _("There are ${count} spots left", mapping={
                    'count': spots
                })


@OrgApp.form(model=FormDefinition, template='form.pt', permission=Public,
             form=lambda self, request: self.form_class)
def handle_defined_form(self, request, form):
    """ Renders the empty form and takes input, even if it's not valid, stores
    it as a pending submission and redirects the user to the view that handles
    pending submissions.

    """

    collection = FormCollection(request.session)

    if not self.current_registration_window:
        spots = 0
        enabled = True
    else:
        spots = 1
        enabled = self.current_registration_window.accepts_submissions(spots)

    if enabled and request.POST:
        submission = collection.submissions.add(
            self.name, form, state='pending', spots=spots)

        return morepath.redirect(request.link(submission))

    layout = FormSubmissionLayout(self, request)

    return {
        'layout': layout,
        'title': self.title,
        'form': enabled and form,
        'definition': self,
        'form_width': 'small',
        'lead': self.meta.get('lead'),
        'text': self.content.get('text'),
        'people': self.people,
        'contact': self.contact_html,
        'coordinates': self.coordinates,
        'hints': tuple(get_hints(layout, self.current_registration_window)),
        'hints_callout': not enabled
    }


@OrgApp.html(model=PendingFormSubmission, template='submission.pt',
             permission=Public, request_method='GET')
@OrgApp.html(model=PendingFormSubmission, template='submission.pt',
             permission=Public, request_method='POST')
@OrgApp.html(model=CompleteFormSubmission, template='submission.pt',
             permission=Private, request_method='GET')
@OrgApp.html(model=CompleteFormSubmission, template='submission.pt',
             permission=Private, request_method='POST')
def handle_pending_submission(self, request):
    """ Renders a pending submission, takes it's input and allows the
    user to turn the submission into a complete submission, once all data
    is valid.

    This view has two states, a completeable state where the form values
    are displayed without a form and an edit state, where a form is rendered
    to change the values.

    Takes the following query parameters for customization::

        * ``edit`` render the view in the edit state
        * ``return-to`` the view redirects to this url once complete
        * ``title`` a custom title (required if external submission)
        * ``quiet`` no success messages are rendered if present

    """
    collection = FormCollection(request.session)

    form = request.get_form(self.form_class, data=self.data)
    form.action = request.link(self)

    if 'edit' not in request.GET:
        form.validate()

    if not request.POST:
        form.ignore_csrf_error()
    elif not form.errors:
        collection.submissions.update(self, form)

    completable = not form.errors and 'edit' not in request.GET

    if completable and 'return-to' in request.GET:

        if 'quiet' not in request.GET:
            request.success(_("Your changes were saved"))

        # the default url should actually never be called
        return request.redirect(request.url)

    if 'title' in request.GET:
        title = request.GET['title']
    else:
        title = self.form.title

    price = get_price(request, form, self)

    # retain some parameters in links (the rest throw away)
    form.action = copy_query(
        request, form.action, ('return-to', 'title', 'quiet'))

    edit_link = URL(copy_query(
        request, request.link(self), ('title', )))

    # the edit link always points to the editable state
    edit_link = edit_link.query_param('edit', '')
    edit_link = edit_link.as_string()

    return {
        'layout': FormSubmissionLayout(self, request, title),
        'title': title,
        'form': form,
        'completable': completable,
        'edit_link': edit_link,
        'complete_link': request.link(self, 'complete'),
        'model': self,
        'price': price,
        'checkout_button': price and request.app.checkout_button(
            button_label=request.translate(_("Pay Online and Complete")),
            title=title,
            price=price,
            email=self.email or self.get_email_field_data(form),
            locale=request.locale
        )
    }


@OrgApp.view(model=PendingFormSubmission, name='complete',
             permission=Public, request_method='POST')
@OrgApp.view(model=CompleteFormSubmission, name='complete',
             permission=Private, request_method='POST')
def handle_complete_submission(self, request):
    form = request.get_form(self.form_class)
    form.process(data=self.data)

    # we're not really using a csrf protected form here (the complete form
    # button is basically just there so we can use a POST instead of a GET)
    form.validate()
    form.ignore_csrf_error()

    if form.errors:
        return morepath.redirect(request.link(self))
    else:
        if self.state == 'complete':
            self.data.changed()  # trigger updates
            request.success(_("Your changes were saved"))

            return morepath.redirect(request.link(
                FormCollection(request.session).scoped_submissions(
                    self.name, ensure_existance=False)
            ))
        else:
            provider = request.app.default_payment_provider
            token = request.params.get('payment_token')

            price = get_price(request, form, self)
            payment = self.process_payment(price, provider, token)

            if not payment:
                request.alert(_("Your payment could not be processed"))
                return morepath.redirect(request.link(self))
            elif payment is not True:
                self.payment = payment

            window = self.registration_window
            if window and not window.accepts_submissions(self.spots):
                request.alert(_("Registrations are no longer possible"))
                return morepath.redirect(request.link(self))

            show_submission = request.params.get('send_by_email') == 'yes'

            self.meta['show_submission'] = show_submission
            self.meta.changed()

            collection = FormCollection(request.session)
            collection.submissions.complete_submission(self)

            # make sure accessing the submission doesn't flush it, because
            # it uses sqlalchemy utils observe, which doesn't like premature
            # flushing at all
            with collection.session.no_autoflush:
                ticket = TicketCollection(request.session).open_ticket(
                    handler_code=self.meta.get('handler_code', 'FRM'),
                    handler_id=self.id.hex
                )
                TicketMessage.create(ticket, request, 'opened')

            send_ticket_mail(
                request=request,
                template='mail_ticket_opened.pt',
                subject=_("Your ticket has been opened"),
                ticket=ticket,
                receivers=(self.email, ),
                content={
                    'model': ticket,
                    'form': form,
                    'show_submission': self.meta['show_submission']
                }
            )

            request.success(_("Thank you for your submission!"))

            return morepath.redirect(request.link(ticket, 'status'))


@OrgApp.view(model=CompleteFormSubmission, name='ticket', permission=Private)
def view_submission_ticket(self, request):
    ticket = TicketCollection(request.session).by_handler_id(self.id.hex)
    return request.redirect(request.link(ticket))


@OrgApp.view(model=CompleteFormSubmission, name='confirm-registration',
             permission=Private, request_method='POST')
def handle_accept_registration(self, request):
    return handle_submission_action(self, request, 'confirmed')


@OrgApp.view(model=CompleteFormSubmission, name='deny-registration',
             permission=Private, request_method='POST')
def handle_deny_registration(self, request):
    return handle_submission_action(self, request, 'denied')


@OrgApp.view(model=CompleteFormSubmission, name='cancel-registration',
             permission=Private, request_method='POST')
def handle_cancel_registration(self, request):
    return handle_submission_action(self, request, 'cancelled')


def handle_submission_action(self, request, action):
    request.assert_valid_csrf_token()

    if action == 'confirmed':
        subject = _("Your registration has been confirmed")
        success = _("The registration has been confirmed")
        failure = _("The registration could not be confirmed")

        def execute():
            if self.registration_window and self.claimed is None:
                return self.claim() or True

    elif action == 'denied':
        subject = _("Your registration has been denied")
        success = _("The registration has been denied")
        failure = _("The registration could not be denied")

        def execute():
            if self.registration_window and self.claimed is None:
                return self.disclaim() or True

    elif action == 'cancelled':
        subject = _("Your registration has been cancelled")
        success = _("The registration has been cancelled")
        failure = _("The registration could not be cancelled")

        def execute():
            if self.registration_window and self.claimed:
                return self.disclaim() or True

    if execute():
        ticket = TicketCollection(request.session).by_handler_id(self.id.hex)

        send_ticket_mail(
            request=request,
            template='mail_registration_action.pt',
            receivers=(self.email, ),
            ticket=ticket,
            content={
                'model': self,
                'action': action,
                'ticket': ticket,
                'form': self.form_obj,
                'show_submission': self.meta.get('show_submission')
            },
            subject=subject
        )

        SubmissionMessage.create(ticket, request, action)

        request.success(success)
    else:
        request.alert(failure)

    return request.redirect(request.link(self))



from cached_property import cached_property
from onegov.chat import Message
from onegov.core.utils import paragraphify, linkify
from onegov.event import Event
from onegov.org import _
from onegov.core.elements import Link, Confirm, Intercooler
from onegov.ticket import Ticket, TicketCollection
from sqlalchemy.orm import object_session

# 👉 when adding new ticket messages be sure to evaluate if they should
# be added to the ticket status page through the org.public_ticket_messages
# setting


class TicketMessageMixin(object):

    def link(self, request):
        return request.class_link(Ticket, {
            'id': self.meta['id'],
            'handler_code': self.meta['handler_code'],
        })

    @cached_property
    def ticket(self):
        return TicketCollection(object_session(self)).by_id(
            self.meta['id'],
            self.meta['handler_code']
        )

    @classmethod
    def create(cls, ticket, request, text=None, owner=None, **extra_meta):
        meta = {
            'id': ticket.id.hex,
            'handler_code': ticket.handler_code,
            'group': ticket.group
        }
        meta.update(extra_meta)

        # force a change of the ticket to make sure that it gets reindexed
        ticket.force_update()

        # the owner can be forced to a specific value
        owner = owner or request.current_username or ticket.ticket_email

        return cls.bound_messages(request.session).add(
            channel_id=ticket.number,
            owner=owner,
            text=text,
            meta=meta
        )


class TicketNote(Message, TicketMessageMixin):
    __mapper_args__ = {
        'polymorphic_identity': 'ticket_note'
    }

    @classmethod
    def create(cls, ticket, request, text, file=None):
        note = super().create(ticket, request, text=text)
        note.file = file

        return note

    @property
    def formatted_text(self):
        return paragraphify(linkify(self.text))

    def links(self, layout):
        yield Link(_("Edit"), layout.request.link(self, 'edit'))
        yield Link(
            _("Delete"), layout.csrf_protected_url(layout.request.link(self)),
            traits=(
                Confirm(
                    _("Do you really want to delete this note?"),
                    _("This cannot be undone."),
                    _("Delete Note"),
                    _("Cancel")
                ),
                Intercooler(
                    request_method="DELETE",
                    redirect_after=layout.request.link(self.ticket)
                )
            ))


class TicketChatMessage(Message, TicketMessageMixin):
    """ Chat messages sent between the person in charge of the ticket and
    the submitter of the ticket.

    Paramters of note:

    - origin: 'external' or 'internal', to differentiate between the
              messages sent from the organisation to someone outside or
              from someone outside to someone inside.

    - notify: only relevant for messages originating from 'internal' - if the
              last sent message with origin 'internal' has this flag, a
              notification is sent to the owner of that message, whenever
              a new external reply comes in.

    """

    __mapper_args__ = {
        'polymorphic_identity': 'ticket_chat'
    }

    @classmethod
    def create(cls, ticket, request, text, owner, origin,
               notify=False, recipient=None):

        return super().create(
            ticket, request, text=text, owner=owner, origin=origin,
            notify=notify, recipient=recipient)

    @property
    def formatted_text(self):
        return paragraphify(linkify(self.text))

    @property
    def subtype(self):
        return self.meta.get('origin', None)


class TicketMessage(Message, TicketMessageMixin):

    __mapper_args__ = {
        'polymorphic_identity': 'ticket'
    }

    @classmethod
    def create(cls, ticket, request, change):
        return super().create(ticket, request, change=change)


class ReservationMessage(Message, TicketMessageMixin):

    __mapper_args__ = {
        'polymorphic_identity': 'reservation'
    }

    @classmethod
    def create(cls, reservations, ticket, request, change):
        return super().create(ticket, request, change=change, reservations=[
            r.id for r in reservations
        ])


class SubmissionMessage(Message, TicketMessageMixin):

    __mapper_args__ = {
        'polymorphic_identity': 'submission'
    }

    @classmethod
    def create(cls, ticket, request, change):
        return super().create(ticket, request, change=change)


class EventMessage(Message, TicketMessageMixin):

    __mapper_args__ = {
        'polymorphic_identity': 'event'
    }

    @classmethod
    def create(cls, event, ticket, request, change):
        return super().create(
            ticket, request, change=change, event_name=event.name)

    def event_link(self, request):
        return request.class_link(Event, {'name': self.meta['event_name']})


class PaymentMessage(Message, TicketMessageMixin):

    __mapper_args__ = {
        'polymorphic_identity': 'payment'
    }

    @classmethod
    def create(cls, payment, ticket, request, change):
        return super().create(
            ticket, request,
            change=change,
            payment_id=payment.id.hex,
            amount=float(payment.amount),
            currency=payment.currency
        )


class DirectoryMessage(Message, TicketMessageMixin):

    __mapper_args__ = {
        'polymorphic_identity': 'directory'
    }

    @classmethod
    def create(cls, directory, ticket, request, action):
        return super().create(
            ticket, request,
            directory_id=directory.id.hex,
            action=action
        )
